package at.ac.uibk.gitsearch.service.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.ResolvableType;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.elasticsearch.repository.ElasticsearchRepository;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * contains various utilities, in order to
 *
 * @author Michael.Breu@uibk.ac.at
 *
 */
@Service
@Transactional
public class UtilityService {

    @SuppressWarnings("unused")
    private final Logger log = LoggerFactory.getLogger(UtilityService.class);

    @Autowired
    private Map<String, JpaRepository<?, ?>> dbRepositories;

    @Autowired
    private Map<String, ElasticsearchRepository<?, ?>> searchRepositories;

    /**
     * method to re-synchronize ES-Entries, if index is not up-to-date. Inspired
     * by
     * https://dev.to/yassinesabri/how-to-synchronize-mysql-database-with-elasticsearch-and-perform-data-querying-in-a-spring-boot-application-k8g
     */
    public List<String> synchronizeElasticSearchFromEntities() {
        List<String> treatedRepositories = new ArrayList<>();
        for (Map.Entry<String, JpaRepository<?, ?>> dbRepositoryEntry : dbRepositories.entrySet()) {
            JpaRepository<?, ?> dbRepository = dbRepositoryEntry.getValue();
            @SuppressWarnings("unchecked")
            final Class<? extends JpaRepository<?, ?>> repositoryClass = (Class<? extends JpaRepository<?, ?>>) dbRepositoryEntry
                .getValue()
                .getClass();

            ResolvableType resolvableType = ResolvableType.forClass(repositoryClass).as(JpaRepository.class);
            ResolvableType entityType = resolvableType.getGeneric(0);
            String dbRepositoryBeanName = dbRepositoryEntry.getKey();
            log.info("Trying to sync Repository {} for {}", dbRepositoryBeanName, entityType.getType().getTypeName());
            String searchRepositoryBeanName = dbRepositoryBeanName.replace("Repository", "SearchRepository");

            ElasticsearchRepository<?, ?> searchRepository = searchRepositories.get(searchRepositoryBeanName);
            if (searchRepository == null) {
                log.debug("Cannot find ES Repository {} for DB Repository {}", dbRepositoryBeanName, searchRepositoryBeanName);
                continue;
            }
            treatedRepositories.add(dbRepositoryBeanName);
            syncEntity(dbRepository, searchRepository);
        }
        return treatedRepositories;
    }

    /**
     * syncs from dataRepository to searchRepository. The sad story is, that it
     * is very hard to work with generic classes.
     *
     * @param dataRepository
     * @param searchRepository
     */
    private void syncEntity(JpaRepository<?, ?> dataRepository, ElasticsearchRepository<?, ?> searchRepository) {
        try {
            Method saveMethod = searchRepository.getClass().getMethod("save", Object.class);
            Pageable nextPage = org.springframework.data.domain.PageRequest.of(0, 50);
            boolean hasNextPage;
            do {
                Page<?> entityList = dataRepository.findAll(nextPage);
                for (Object entity : entityList.toList()) {
                    saveMethod.invoke(searchRepository, entity);
                }
                hasNextPage = entityList.hasNext();
                if (hasNextPage) {
                    nextPage = entityList.nextPageable();
                }
            } while (hasNextPage);
        } catch (
            NoSuchMethodException | SecurityException | IllegalAccessException | IllegalArgumentException | InvocationTargetException e
        ) {
            log.error("Cannot sync {} to {}", dataRepository, searchRepository, e);
        }
    }
}
